/* FFT Windowing
 * Demonstration code by Emil Mikulic, 2002.
 * http://dmr.ath.cx/sound/window/
 *
 * Feel free to do whatever you like with this code.
 * Feel free to credit me.
 */

#include <math.h>
#include <stdio.h>

/* http://www.zip.com.au/~erikd/libsndfile/ */
#include <sndfile.h>

#ifndef PI
# ifdef M_PI
#  define PI M_PI
# else
/* overkill */
#  define PI 3.141592653589793238462643383279502884197169399375105820974
# endif
#endif

#define POWER		16
#define SIZE		(1<<POWER)
#define INFILE		"wndchunk.wav"
#define WINDOW_FUNC	w_blackman

double i_re[SIZE], i_im[SIZE], out_amp[SIZE/2+1];
SF_INFO sfinfo;
SNDFILE *sndfile;



/* --- error handling --- */
__inline void ERR(int condition, char *reason)
{
	if (!condition) return;

	/* on condition, throw reason */
	fprintf(stderr, "%s\n", reason);
	exit(1);
}



/* --- windowing --- */
__inline double w_hann(const double pos)
{
	double x = cos(pos*PI/2.0);
	return x*x;
}

__inline double w_rect(const double pos)
{
	return 1.0;
}

__inline double w_hamming(const double pos)
{
	return 0.54 + 0.46 * cos(PI*pos);
}

__inline double w_blackman(const double pos)
{
	return 0.42 + 0.5 * cos(PI*pos) + 0.08 * cos(2.0*PI*pos);
}

__inline double w_gaussian(const double pos)
{
	double a = 4.0;
	double ax = a * pos;
	return exp(-0.5 * ax*ax);
}

__inline double w_welch(const double pos)
{
	return 1.0 - pos*pos;
}

__inline double w_bartlett(const double pos)
{
	if (pos == 0.0)
		return 1.0;
	else if (pos > -1.0 && pos < 0.0)
		return pos + 1.0;
	else if (pos > 0.0 && pos < 1.0)
		return 1.0 - pos;
	else
		return 0.0;
}



/* --- FFT --- */
__inline int br(const int i, const int power)
{
	int j, k = 0;

	for (j=0; j<power; j++)
	if ((i >> j) & 1)
		k = k | (1<<(power-j-1));

	return k;
}

__inline void fft(double *fft_re, double *fft_im, 
	const double sign, const int fft_size)
{
	int blocks, points, i, power;

	power = (int)(log((double)fft_size) / log(2.0));
	ERR((1<<power != fft_size), "FFT_SIZE not a power of two!");

	/* bit reverse order */
	for (i=0; i<fft_size; i++)
	{
		int k = br(i, power);
		if (i > k)
		{
			double t;

			t = fft_re[i];
			fft_re[i] = fft_re[k];
			fft_re[k] = t;

			t = fft_im[i];
			fft_im[i] = fft_im[k];
			fft_im[k] = t;
		}
	}

	/* pfft! */
	for (blocks=fft_size/2,points=2; blocks>=1; blocks/=2,points*=2)
	for (i=0; i<blocks; i++)
	{
		int j, offset = points * i;

		/* "butterfly" odd elements */
		for (j=0; j<points/2; j++)
		{
			double t_re, t_im, top_r, top_i, btm_r, btm_i, theta;
			theta = sign * 2.0 * PI * (double)j / (double)points;
			t_re = cos(theta);
			t_im = sin(theta);

			top_r = fft_re[j+offset];
			top_i = fft_im[j+offset];

			btm_r = fft_re[j+points/2+offset]*t_re -
				fft_im[j+points/2+offset]*t_im;

			btm_i = fft_re[j+points/2+offset]*t_im +
				fft_im[j+points/2+offset]*t_re;

			fft_re[j+offset] = top_r + btm_r;
			fft_im[j+offset] = top_i + btm_i;

			fft_re[j+points/2+offset] = top_r - btm_r;
			fft_im[j+points/2+offset] = top_i - btm_i;
		}
	}
}



int main()
{
	int i;
	double loudest = 0.0;

	sndfile = sf_open_read(INFILE, &sfinfo);
	ERR(!sndfile, "Can't open file for reading");

	/* Informational */
	fprintf(stderr, "Opened sample --\n\n"
		"samplerate:  %d\n"
		"samples:     %d\n"
		"channels:    %d\n"
		"pcmbitwidth: %d\n\n",
		sfinfo.samplerate, sfinfo.samples,
		sfinfo.channels, sfinfo.pcmbitwidth);

	ERR(sfinfo.channels != 1, "Need mono input");

	/* Fill up input */
	ERR(sf_read_double(sndfile, i_re, SIZE, 1) != SIZE,
		"Failed to read samples");

	/* Window input and zero out imaginaries */
	for (i=0; i<SIZE; i++)
	{
		i_re[i] *= WINDOW_FUNC((double)(i-SIZE/2) / (double)(SIZE/2));
		i_im[i] = 0.0;
	}

	/* yay */
	fft(i_re, i_im, -1.0, SIZE);

	/* Amplitudes from complex */
	for (i=0; i<=SIZE/2; i++)
	{
		double re = i_re[i], im = i_im[i];

		out_amp[i] = sqrt(re*re + im*im) / (double)SIZE;
		if (out_amp[i] > loudest) loudest = out_amp[i];
	}

	for (i=0; i<=SIZE/2; i++)
		printf("%f %f\n", (double)(i*sfinfo.samplerate)/(double)SIZE,
			out_amp[i] / loudest);

	return 0;
}
